# BlenderBIM Add-on - OpenBIM Blender Add-on
# Copyright (C) 2020, 2021, 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of BlenderBIM Add-on.
#
# BlenderBIM Add-on is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BlenderBIM Add-on is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BlenderBIM Add-on.  If not, see <http://www.gnu.org/licenses/>.

import bpy
import ifcopenshell.api
import blenderbim.bim.helper
import blenderbim.tool as tool
import blenderbim.bim.module.model.profile as model_profile
import blenderbim.core.profile as core
from blenderbim.bim.module.model.decorator import ProfileDecorator
from blenderbim.bim.module.profile.prop import generate_thumbnail_for_active_profile
from blenderbim.bim.module.profile.data import refresh


class LoadProfiles(bpy.types.Operator):
    bl_idname = "bim.load_profiles"
    bl_label = "Load Profiles"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        props = context.scene.BIMProfileProperties
        props.profiles.clear()

        for profile in tool.Ifc.get().by_type("IfcProfileDef"):
            if not profile.ProfileName:
                continue
            new = props.profiles.add()
            new.ifc_definition_id = profile.id()
            new.name = profile.ProfileName or "Unnamed"
            new.ifc_class = profile.is_a()

        props.is_editing = True
        bpy.ops.bim.disable_editing_profile()
        return {"FINISHED"}


class DisableProfileEditingUI(bpy.types.Operator):
    bl_idname = "bim.disable_profile_editing_ui"
    bl_label = "Disable Profile Editing UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.BIMProfileProperties.is_editing = False
        return {"FINISHED"}


class RemoveProfileDef(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.remove_profile_def"
    bl_label = "Remove Profile Definition"
    bl_options = {"REGISTER", "UNDO"}
    profile: bpy.props.IntProperty()

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        current_index = props.active_profile_index
        ifcopenshell.api.run("profile.remove_profile", tool.Ifc.get(), profile=tool.Ifc.get().by_id(self.profile))
        bpy.ops.bim.load_profiles()

        # preserve selected index if possible
        if props.profiles:
            props.active_profile_index = min(current_index, len(props.profiles) - 1)


class EnableEditingProfile(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.enable_editing_profile"
    bl_label = "Enable Editing Profile"
    bl_options = {"REGISTER", "UNDO"}
    profile: bpy.props.IntProperty()

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        props.profile_attributes.clear()
        blenderbim.bim.helper.import_attributes2(tool.Ifc.get().by_id(self.profile), props.profile_attributes)
        props.active_profile_id = self.profile


class DisableEditingProfile(bpy.types.Operator):
    bl_idname = "bim.disable_editing_profile"
    bl_label = "Disable Editing Profile"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.BIMProfileProperties.active_profile_id = 0
        bpy.ops.bim.disable_editing_arbitrary_profile()
        return {"FINISHED"}


class EditProfile(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.edit_profile"
    bl_label = "Edit Profile"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        attributes = blenderbim.bim.helper.export_attributes(props.profile_attributes)
        profile = tool.Ifc.get().by_id(props.active_profile_id)
        ifcopenshell.api.run("profile.edit_profile", tool.Ifc.get(), profile=profile, attributes=attributes)
        model_profile.DumbProfileRegenerator().regenerate_from_profile_def(profile)
        bpy.ops.bim.load_profiles()
        generate_thumbnail_for_active_profile()


class AddProfileDef(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.add_profile_def"
    bl_label = "Add Profile"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        profile_class = props.profile_classes
        if profile_class == "IfcArbitraryClosedProfileDef":
            points = [(0, 0), (0.1, 0), (0.1, 0.1), (0, 0.1), (0, 0)]
            profile = ifcopenshell.api.run("profile.add_arbitrary_profile", tool.Ifc.get(), profile=points)
        else:
            profile = ifcopenshell.api.run("profile.add_parameterized_profile", tool.Ifc.get(), ifc_class=profile_class)
        profile.ProfileName = "New Profile"
        bpy.ops.bim.load_profiles()


class DuplicateProfileDef(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.duplicate_profile_def"
    bl_label = "Duplicate Profile"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        props = context.scene.BIMProfileProperties
        if len(props.profiles) > props.active_profile_index:
            return True
        cls.poll_message_set("No profile selected to duplicate.")
        return False

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        ifc_file = tool.Ifc.get()
        profile = ifc_file.by_id(props.profiles[props.active_profile_index].ifc_definition_id)
        tool.Profile.duplicate_profile(profile)
        bpy.ops.bim.load_profiles()


class EnableEditingArbitraryProfile(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.enable_editing_arbitrary_profile"
    bl_label = "Enable Editing Arbitrary Profile"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        active_profile = props.profiles[props.active_profile_index]
        profile_id = active_profile.ifc_definition_id
        props.active_arbitrary_profile_id = profile_id
        profile = tool.Ifc.get().by_id(profile_id)
        obj = tool.Model.import_profile(profile)
        tool.Ifc.link(profile, obj)
        bpy.context.scene.collection.objects.link(obj)
        tool.Blender.select_and_activate_single_object(context, obj)
        bpy.ops.object.mode_set(mode="EDIT")
        ProfileDecorator.install(context, exit_edit_mode_callback=lambda: disable_editing_arbitrary_profile(context))
        tool.Blender.set_viewport_tool("bim.cad_tool")


def disable_editing_arbitrary_profile(context):
    obj = context.active_object
    if obj and obj.type == "MESH" and obj.data and obj.data.BIMMeshProperties.subshape_type == "PROFILE":
        ProfileDecorator.uninstall()
        bpy.ops.object.mode_set(mode="OBJECT")
        profile_mesh = obj.data
        bpy.data.objects.remove(obj)
        bpy.data.meshes.remove(profile_mesh)

    props = context.scene.BIMProfileProperties
    props.active_arbitrary_profile_id = 0
    # need to update profile manager ui
    # if this was called from decorator
    refresh()


class DisableEditingArbitraryProfile(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.disable_editing_arbitrary_profile"
    bl_label = "Disable Editing Arbitrary Profile"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        return disable_editing_arbitrary_profile(context)


class EditArbitraryProfile(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.edit_arbitrary_profile"
    bl_label = "Edit Arbitrary Profile"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        old_profile = tool.Ifc.get().by_id(props.active_arbitrary_profile_id)

        obj = context.active_object

        ProfileDecorator.uninstall()
        bpy.ops.object.mode_set(mode="OBJECT")

        profile = tool.Model.export_profile(obj)
        if not profile:

            def msg(self, context):
                self.layout.label(text="INVALID PROFILE: " + indices[1])

            bpy.context.window_manager.popup_menu(msg, title="Error", icon="ERROR")
            ProfileDecorator.install(
                context, exit_edit_mode_callback=lambda: disable_editing_arbitrary_profile(context)
            )
            bpy.ops.object.mode_set(mode="EDIT")
            return

        prev_profile_id = profile.id()
        profile_mesh = obj.data
        bpy.data.objects.remove(obj)
        bpy.data.meshes.remove(profile_mesh)

        profile.ProfileType = old_profile.ProfileType
        profile.ProfileName = old_profile.ProfileName
        for inverse in tool.Ifc.get().get_inverse(old_profile):
            ifcopenshell.util.element.replace_attribute(inverse, old_profile, profile)
        ifcopenshell.util.element.remove_deep2(tool.Ifc.get(), old_profile)
        bpy.ops.bim.load_profiles()
        if props.active_profile_id == prev_profile_id:
            props.active_profile_id = profile.id()
        props.active_arbitrary_profile_id = 0

        model_profile.DumbProfileRegenerator().regenerate_from_profile_def(profile)


class PurgeUnusedProfiles(bpy.types.Operator, tool.Ifc.Operator):
    bl_idname = "bim.purge_unused_profiles"
    bl_label = "Purge Unused Profiles"
    bl_options = {"REGISTER", "UNDO"}

    def _execute(self, context):
        props = context.scene.BIMProfileProperties
        purged_profiles = core.purge_unused_profiles(tool.Ifc, tool.Profile)
        self.report({"INFO"}, f"{purged_profiles} profiles were purged.")

        if props.is_editing:
            refresh()
            bpy.ops.bim.load_profiles()
